/*
    UHU-menu
    Copyright (C) 2003  Koblinger Egmont <egmont@uhulinux.hu>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#define _GNU_SOURCE
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <locale.h>
#include <langinfo.h>
#include <dlfcn.h>
#include <dirent.h>
#include <getopt.h>
#include <errno.h>
#include <sys/stat.h>

#define UHU_MENU_CORE_ENGINE
#include "uhu-menu.h"

#define MENU_SRC_DIR "/usr/share"
#define MENU_DST_DIR "/usr/lib/menu"
#define PLUGIN_DIR "/usr/lib/uhu-menu"
#define INPUT_PLUGIN "desktop"

const char *menu_src_dir = MENU_SRC_DIR;
const char *menu_dst_dir = MENU_DST_DIR;
const char *plugin_dir = PLUGIN_DIR;
const char *input_plugin = INPUT_PLUGIN;
char input_plugin_dir[PATH_MAX];
char output_plugin_dir[PATH_MAX];

#define LOGLEVEL 5
int loglevel = LOGLEVEL;

char buf[PATH_MAX];

char *wm_only = NULL;
char *wm_skip = NULL;

char *lang_only = NULL;
char *lang_add = NULL;
char *lang_skip = NULL;

struct language *current_language;

int plugin_type = -1;
const char *plugin_name = NULL;

#define PLUGINS_MAX 64
struct plugin {
	char *name;
	int type;
	int ok;
	void *handle;
	struct item *(*read_menu) (const char *path);
	int (*write_menu) (const struct item *item, const char *path, struct keep_files *keep_files);
	int (*write_menu_lang) (const struct item *item, const char *path, const struct language *language, struct keep_files *keep_files);
	int *need_all_entries;
};

struct plugin iplugin;
struct plugin oplugins[PLUGINS_MAX];
struct keep_files keep_files_array[PLUGINS_MAX];
int opluginnr = 0;

int plugin_sort_fn (const void *v1, const void *v2)
{
	const struct plugin *p1 = v1;
	const struct plugin *p2 = v2;
	return strcmp(p1->name, p2->name);
}

/*
 *  qsort_menu + menu_sort_fn
 *  Sorts the menu according to current_lang. Invalid entries are put to the
 *  end of the list. "nr" is set to the number of valid entries, as opposed to
 *  nr_real which counts the invalid entries too.
 */
int menu_sort_fn (const void *v1, const void *v2)
{
	const char *compare_fields[] = { "Name", "GenericName", "Comment", "Exec", "TryExec", "Icon", "Categories", NULL };
	int compare_nr;
	const struct item *i1 = v1;
	const struct item *i2 = v2;
	const struct key *k1, *k2;
	int retval;
	if (i1->valid != i2->valid) return i2->valid - i1->valid;
	if (i1->issubmenu != i2->issubmenu) return i2->issubmenu - i1->issubmenu;
	if (i1->major != i2->major) return i2->major - i1->major;
	if (i1->minor != i2->minor) return i2->minor - i1->minor;
	for (compare_nr = 0; compare_fields[compare_nr] != NULL; compare_nr++) {
		k1 = item_find_key(i1, NULL, compare_fields[compare_nr]);
		k2 = item_find_key(i2, NULL, compare_fields[compare_nr]);
		if (k1 == NULL && k2 == NULL) continue;
		if (k1 == NULL) return -1;
		if (k2 == NULL) return 1;
		if (k1->current_lang == NULL && k2->current_lang == NULL) continue;
		if (k1->current_lang == NULL) return -1;
		if (k2->current_lang == NULL) return 1;
		if (current_language->locale != NULL) {
			retval = strcoll_l(k1->current_lang->value, k2->current_lang->value, current_language->locale);
		} else {
			retval = strcmp(k1->current_lang->value, k2->current_lang->value);
		}
		if (retval != 0) return retval;
	}
	return 0;
}

void qsort_subitems (struct item *item)
{
	int i;
	qsort(item->subitems, item->subitems_nr_real, sizeof(struct item), menu_sort_fn);
	for (i = 0; i < item->subitems_nr_real && item->subitems[i].valid; i++) {
		if (item->subitems[i].issubmenu) {
			qsort_subitems(&item->subitems[i]);
		}
	}
	item->subitems_nr = i;
}

/*
 *  empty_is_invalid
 *  Marks empty submenus (those which do not have valid and nonempty descendants)
 *  as invalid.
 */
int empty_is_invalid (struct item *item)
{
	int i;
	if (!item->valid) return -1;
	if (!item->issubmenu) return 0;
	item->valid = 0;
	for (i = 0; i < item->subitems_nr_real; i++) {
		if (empty_is_invalid(&item->subitems[i]) == 0) {
			item->valid = 1;
		}
	}
	return item->valid ? 0 : -1;
}

/*
 *  prepare_for_wm
 *  Prepares the menu for a particular window manager. Marks entries that
 *  should be skipped for this wm as invalid.
 */
int prepare_for_wm (struct item *item, const char *wm, int force_valid)
{
	int i;

	item->valid_wm = 1;
	if (!force_valid) {
		if (item->onlyshowin != NULL && !is_in_list(wm, item->onlyshowin, ",;")) {
			item->valid_wm = 0;
		}
		if (item->notshowin != NULL && is_in_list(wm, item->notshowin, ",;")) {
			item->valid_wm = 0;
		}
	}
	item->valid = item->valid_general && item->valid_wm && item->valid_lang;

	if (item->issubmenu) {
		for (i = 0; i < item->subitems_nr_real; i++) {
			prepare_for_wm(&item->subitems[i], wm, force_valid);
		}
	}
	return 0;
}

/*
 *  prepare_for_lang
 *  Prepares the menu for a particular language. Sets the current_lang pointer
 *  in every key of every item for faster access. Converts major and minor
 *  entries from string to int. Does not yet sort the entries.
 */
int prepare_for_lang (struct item *item, const struct language *language, int force_valid)
{
	int i;
	struct section *s;
	struct key *k;

	s = item->sections;
	while (s != NULL) {
		k = s->keys;
		while (k != NULL) {
			k->current_lang = key_find_lang(k, language->name);
			k = k->next;
		}
		s = s->next;
	}
	item->valid_lang = 1;
	if (!force_valid) {
		if (item->key_name == NULL || item->key_name->current_lang == NULL) {
			item->valid_lang = 0;
		}
	}
	if (item->key_major == NULL || item->key_major->current_lang == NULL) {
		item->major = 0;
	} else {
		item->major = string_to_int(item->key_major->current_lang->value, 0);
	}
	if (item->key_minor == NULL || item->key_minor->current_lang == NULL) {
		item->minor = 0;
	} else {
		item->minor = string_to_int(item->key_minor->current_lang->value, 0);
	}
	item->valid = item->valid_general && item->valid_wm && item->valid_lang;

	if (item->issubmenu) {
		for (i = 0; i < item->subitems_nr_real; i++) {
			prepare_for_lang(&item->subitems[i], language, force_valid);
		}
	}
	return 0;
}

/*
 *  fix_menu
 *  Called only once after the input plugin has built up the menu tree and
 *  before the output plugins dump it. Fills out some fields which might yet
 *  be uninitialized but may be needed by this core engine or an output plugin.
 */
int fix_menu (struct item *item)
{
	int i;

	item->key_name  = item_find_key(item, "", "Name");
	item->key_major = item_find_key(item, "", "X-UHU-Major");
	item->key_minor = item_find_key(item, "", "X-UHU-Minor");
	item->valid_general =
	  (item->sections != NULL &&
	   get_value_bool(item, NULL, "Hidden", NULL) != 1 &&
	   get_value_bool(item, NULL, "NoDisplay", NULL) != 1);
	item->terminal = (get_value_bool(item, NULL, "Terminal", NULL) == 1);
	item->onlyshowin = get_value(item, NULL, "OnlyShowIn", NULL);
	item->notshowin = get_value(item, NULL, "NotShowIn", NULL);

	if (item->issubmenu) {
		item->subitems_nr_real = item->subitems_nr;
		for (i = 0; i < item->subitems_nr_real; i++) {
			fix_menu(&item->subitems[i]);
		}
	}
	return 0;
}

int output_all_in_one (struct item *item)
{
	int i, j;

	prepare_for_lang(item, current_language, 1);

	for (i = 0; i < opluginnr; i++) {
		if (!oplugins[i].ok) continue;
		if (oplugins[i].write_menu == NULL) continue;

		prepare_for_wm(item, oplugins[i].name, oplugins[i].need_all_entries != NULL ? 1 : 0);
		empty_is_invalid(item);
		qsort_subitems(item);

		sprintf(buf, "%s/%s", menu_dst_dir, oplugins[i].name);
		mkdir(buf, 0777);

		message(6, "Invoking write_menu() of output plugin `%s'", oplugins[i].name);
		plugin_type = 1;
		plugin_name = oplugins[i].name;
		j = (*oplugins[i].write_menu)(item, buf, &keep_files_array[i]);
		plugin_type = -1;
		plugin_name = NULL;
		if (j >= 0) message(6, "write_menu() returned successfully");
		else message(6, "write_menu() returned an error");
	}
	return 0;
}

int output_one_lang (struct item *item)
{
	int i, j;

	if (current_language->locale == NULL) return -1;

	prepare_for_lang(item, current_language, 0);
	for (i = 0; i < opluginnr; i++) {
		if (!oplugins[i].ok) continue;
		if (oplugins[i].write_menu_lang == NULL) continue;

		prepare_for_wm(item, oplugins[i].name, oplugins[i].need_all_entries != NULL ? 1 : 0);
		empty_is_invalid(item);
		qsort_subitems(item);

		sprintf(buf, "%s/%s", menu_dst_dir, oplugins[i].name);
		mkdir(buf, 0777);

		message(6, "Invoking write_menu_lang() of output plugin `%s' for language `%s'", oplugins[i].name, current_language->name);
		plugin_type = 1;
		plugin_name = oplugins[i].name;
		j = (*oplugins[i].write_menu_lang)(item, buf, current_language, &keep_files_array[i]);
		plugin_type = -1;
		plugin_name = NULL;
		if (j >= 0) message(6, "write_menu_lang() returned successfully");
		else message(6, "write_menu_lang() returned an error");
	}
	return 0;
}

void keep_process_dir (struct keep_files *keep_files, const char *dirname)
{
	DIR *dir;
	struct dirent *de;
	keep_files->dirname = strdup(dirname);
	keep_files->allocated = 8;
	keep_files->filenames = malloc(keep_files->allocated * sizeof(char *));
	keep_files->keep_flags = malloc(keep_files->allocated * sizeof(bool));
	keep_files->count = 0;
	dir = opendir(dirname);
	if (dir == NULL) return;
	while ((de = readdir(dir)) != NULL) {
		if (!strcmp(de->d_name, ".")) continue;
		if (!strcmp(de->d_name, "..")) continue;
		if (keep_files->count == keep_files->allocated) {
			keep_files->allocated *= 2;
			keep_files->filenames = realloc(keep_files->filenames, keep_files->allocated * sizeof(char *));
			keep_files->keep_flags = realloc(keep_files->keep_flags, keep_files->allocated * sizeof(bool));
		}
		keep_files->filenames[keep_files->count] = strdup(de->d_name);
		keep_files->keep_flags[keep_files->count] = 0;
		keep_files->count++;
	}
}

void keep_perform_cleanup (struct keep_files *keep_files)
{
	int i;
	char buf[PATH_MAX];
	for (i = 0; i < keep_files->count; i++) {
		if (!keep_files->keep_flags[i]) {
			sprintf(buf, "%s/%s", keep_files->dirname, keep_files->filenames[i]);
			rm_rf(buf);
		}
	}
}

void list ()
{
	int i;
	for (i = 0; i < opluginnr; i++) {
		if (!oplugins[i].ok) continue;
		printf("%-15s (%s)\n",
		  oplugins[i].name,
		  oplugins[i].write_menu_lang != NULL ? "separate config for each language"
		    : "one config for all languages");
	}
}

void help ()
{
	printf(
	  "Usage: uhu-menu [options...]\n"
	  "\n"
	  "  -s, --source-dir       Source dir of menu (default: " MENU_SRC_DIR ")\n"
	  "  -d, --destination-dir  Destination dir of menu (default: " MENU_DST_DIR ")\n"
	  "  -p, --plugin-dir       Plugin dir (default: " PLUGIN_DIR ")\n"
	  "  -i, --input-plugin     Input plugin (default: " INPUT_PLUGIN ")\n"
	  "      --list-plugins     List available window manager plugins\n"
	  "      --wm-only          Process only these window managers (comma separated)\n"
	  "      --wm-skip          Skip these window managers (comma separated)\n"
	  "      --lang-only        Process only these languages (comma separated)\n"
	  "      --lang-add         Make sure to process these languages (comma separated)\n"
	  "      --lang-skip        Skip these languages (comma separated)\n"
	  "  -l, --loglevel         Loglevel (0 - 8) (default: %d)\n"
	  "  -h, --help             Guess what :-)\n"
	, LOGLEVEL);
}

int main (int argc, char *argv[])
{
	struct item *item;
	const char *err1, *err2;
	DIR *d;
	struct dirent *de;
	int c;
	int l;
	int i;
	int opt_list = 0;
	char *s;

	static const struct option long_options[] = {
		{ "source-dir",      1, NULL, 's' },
		{ "src-dir",         1, NULL, 's' },
		{ "destination-dir", 1, NULL, 'd' },
		{ "dest-dir",        1, NULL, 'd' },
		{ "dst-dir",         1, NULL, 'd' },
		{ "plugin-dir",      1, NULL, 'p' },
		{ "input-plugin",    1, NULL, 'i' },
		{ "list-plugins",    0, NULL, 256 },
		{ "wm-only",         1, NULL, 257 },
		{ "wm-skip",         1, NULL, 258 },
		{ "lang-only",       1, NULL, 259 },
		{ "lang-add",        1, NULL, 260 },
		{ "lang-skip",       1, NULL, 261 },
		{ "loglevel",        1, NULL, 'l' },
		{ "help",            0, NULL, 'h' },
		{ NULL,              0, NULL, 0   }
	};

	/* Initialization */

	setlocale(LC_ALL, "");

	while ((c = getopt_long(argc, argv, "s:d:p:i:l:h", long_options, NULL)) != -1) {
		switch (c) {
		  case 's':
			menu_src_dir = optarg;
			break;
		  case 'd':
			menu_dst_dir = optarg;
			break;
		  case 'p':
			plugin_dir = optarg;
			break;
		  case 'i':
			input_plugin = optarg;
			break;
		  case 256:
			opt_list = 1;
			break;
		  case 257:
		  case 258:
			if (wm_only != NULL || wm_skip != NULL) {
				message(0, "Window manager list already specified");
				exit(1);
			}
			if (c == 257) wm_only = optarg;
			else wm_skip = optarg;
			break;
		  case 259:
			if (lang_only != NULL || lang_add != NULL || lang_skip != NULL) {
				message(0, "Language list already specified");
				exit(1);
			}
			lang_only = optarg;
			break;
		  case 260:
			if (lang_only != NULL || lang_add != NULL) {
				message(0, "Language list already specified");
				exit(1);
			}
			lang_add = optarg;
			break;
		  case 261:
			if (lang_only != NULL || lang_skip != NULL) {
				message(0, "Language list already specified");
				exit(1);
			}
			lang_skip = optarg;
			break;
		  case 'l':
			loglevel = string_to_int(optarg, -1);
			if (loglevel < 0 || loglevel > 8) {
				message(0, "Invalid loglevel");
				exit(1);
			};
			break;
		  case 'h':
			help();
			exit(0);
		  case '?':
		  case ':':
			exit(1);
		}
	}
	sprintf(input_plugin_dir, "%s/plugins/input", plugin_dir);
	sprintf(output_plugin_dir, "%s/plugins/output", plugin_dir);

	find_or_create_language("");

	/* Scan output plugin candidates */

	d = opendir(output_plugin_dir);
	if (d == NULL) {
		message_e(0, "Cannot open plugin dir %s", output_plugin_dir);
		exit(1);
	}
	while ((de = readdir(d)) != NULL) {
		if (!strcmp(de->d_name, "." )) continue;
		if (!strcmp(de->d_name, "..")) continue;
		l = strlen(de->d_name);
		if (l < 4
		  || strcmp(de->d_name + l - 3, ".so")) {
			message(4, "Skipping non-plugin %s/%s", output_plugin_dir, de->d_name);
			continue;
		}
		if (opluginnr == PLUGINS_MAX) {
			message(0, "Too many output plugins found");
			exit(1);
		}
		sprintf(buf, "%.*s", l - 3, de->d_name);
		if (wm_only != NULL && !is_in_list(buf, wm_only, ",;")) continue;
		if (wm_skip != NULL &&  is_in_list(buf, wm_skip, ",;")) continue;
		oplugins[opluginnr].name = strdup(buf);
		oplugins[opluginnr].ok = 1;
		opluginnr++;
	}
	closedir(d);
	qsort(oplugins, opluginnr, sizeof(struct plugin), plugin_sort_fn);

	/* Open output plugins */

	for (i = 0; i < opluginnr; i++) {
		sprintf(buf, "%s/%s.so", output_plugin_dir, oplugins[i].name);
		oplugins[i].handle = dlopen(buf, RTLD_NOW);
		err1 = dlerror();
		if (err1 != NULL) {
			message(3, "dlopen %s: %s", buf, err1);
			continue;
		}
		oplugins[i].write_menu = dlsym(oplugins[i].handle, "write_menu");
		err1 = dlerror();
		oplugins[i].write_menu_lang = dlsym(oplugins[i].handle, "write_menu_lang");
		err2 = dlerror();
		if ((err1 == NULL && err2 == NULL) || (err1 != NULL && err2 != NULL)) {
			oplugins[i].ok = 0;
			message(3, "Ignoring faulty output plugin %s", oplugins[i].name);
			dlclose(oplugins[i].handle);
			continue;
		}
		oplugins[i].need_all_entries = dlsym(oplugins[i].handle, "need_all_entries");
	}

	if (opt_list) {
		list();
		exit(0);
	}

	/* Open input plugin */

	iplugin.name = strdup(input_plugin);
	sprintf(buf, "%s/%s.so", input_plugin_dir, input_plugin);
	iplugin.handle = dlopen(buf, RTLD_NOW);
	err1 = dlerror();
	if (err1 != NULL) {
		message(0, "dlopen %s: %s", buf, err1);
		exit(1);
	}
	iplugin.read_menu = dlsym(iplugin.handle, "read_menu");
	err1 = dlerror();
	if (err1 != NULL) {
		message(0, "Faulty input plugin!");
		exit(1);
	}

	/* Read the menu */

	message(6, "Invoking read_menu() of input plugin `%s'", iplugin.name);
	plugin_type = 0;
	plugin_name = iplugin.name;
	item = (*iplugin.read_menu)(menu_src_dir);
	plugin_type = -1;
	plugin_name = NULL;
	if (item == NULL) {
		message(0, "Could not read the menu!");
		exit(1);
	}
	else message(6, "Input plugin successfully returned");

	/* Fixing the menu (filling some fields etc.) */

	fix_menu(item);

	/* Get the current directory listings */

	for (i = 0; i < opluginnr; i++) {
		if (!oplugins[i].ok) continue;
		sprintf(buf, "%s/%s", menu_dst_dir, oplugins[i].name);
		keep_process_dir(&keep_files_array[i], buf);
	}

	/* Process all-in-one output plugins */

	current_language = find_or_create_language("");
	output_all_in_one(item);

	/* Process once-for-each-language plugins */

	current_language = find_or_create_language("");
	output_one_lang(item);

	s = lang_only;
	if (s == NULL) s = lang_add;
	if (s != NULL) {
		char *lang;
		while (*s != '\0') {
			s += strspn(s, ",;");
			if (*s == '\0') break;
			i = strcspn(s, ",;");
			lang = malloc(i + 1);
			strncpy(lang, s, i);
			lang[i] = '\0';
			s += i;
			current_language = find_or_create_language(lang);
			free(lang);
			output_one_lang(item);
		}
	}

	if (lang_only == NULL) {
		current_language = first_language;
		while (current_language != NULL) {
			if (current_language->name[0] != '\0' &&
			  !is_in_list(current_language->name, lang_add, ",;") &&
			  !is_in_list(current_language->name, lang_skip, ",;"))
				output_one_lang(item);
			current_language = current_language->next;
		}
	}

	/* Remove files that weren't touched by the plugins */

	for (i = 0; i < opluginnr; i++) {
		if (oplugins[i].ok) keep_perform_cleanup(&keep_files_array[i]);
	}

	/* Close output plugins and input plugin */

	for (i = 0; i < opluginnr; i++) {
		if (oplugins[i].ok) dlclose(oplugins[i].handle);
	}
	dlclose(iplugin.handle);

	/* Yippie! Survived without segfault :-) */

	return 0;
}
